import argparse
import json
import jsonschema
import logging
import os
import requests
import sys
import traceback
import yaml
from flasgger import Swagger
from flask import Flask, jsonify, g, Response, request
from flask_restful import Api, Resource
from flask_cors import CORS
from dug.core import Search

"""
Defines the semantic search API

This exists in large part because it's not safe to expose the Elasticsearch interface to the internet.
So we'll need to add validation to ensure reasonable reasonable, well formed, valid requests inbound.
"""
logger = logging.getLogger (__name__)

app = Flask(__name__)

""" Enable CORS. """
api = Api(app)
CORS(app)
debug=False

""" Load the schema. """
schema_file_path = os.path.join (os.path.dirname (__file__), 'conf', 'api-schema.yaml')
template = None
with open(schema_file_path, 'r') as file_obj:
    template = yaml.load(file_obj, Loader=yaml.FullLoader)

""" Describe the API. """
app.config['SWAGGER'] = {
    'title': 'Dug Search API',
    'description': 'An API, compiler, and executor for cloud native distributed systems.',
    'uiversion': 3
}

swagger = Swagger(app, template=template)

def dug ():
    if not hasattr(g, 'dug'):
        g.search = Search ()
    return g.search
    
class DugResource(Resource):
    """ Base class handler for Dug API requests. """
    def __init__(self):
        self.specs = {}
        
    """ Functionality common to Dug services. """
    def validate (self, request, component):
        return
        """ Validate a request against the schema. """
        if not self.specs:
            with open(schema_file_path, 'r') as file_obj:
                self.specs = yaml.load(file_obj)
        to_validate = self.specs["components"]["schemas"][component]
        try:
            app.logger.debug (f"--:Validating obj {json.dumps(request.json, indent=2)}")
            app.logger.debug (f"  schema: {json.dumps(to_validate, indent=2)}")            
            jsonschema.validate(request.json, to_validate)
        except jsonschema.exceptions.ValidationError as error:
            app.logger.error (f"ERROR: {str(error)}")
            traceback.print_exc (error)
            abort(Response(str(error), 400))

    def create_response (self, result=None, status='success', message='', exception=None):
        """ Create a response. Handle formatting and modifiation of status for exceptions. """
        if exception:
            traceback.print_exc ()
            status='error'
            exc_type, exc_value, exc_traceback = sys.exc_info()
            result = {
                'error' : repr(traceback.format_exception(exc_type, exc_value, exc_traceback))
            }
        return {
            'status'  : status,
            'result'  : result,
            'message' : message
        }

class DugSearchResource(DugResource):
    """ Execute a search """

    """ System initiation. """
    def post(self):
        """
        Execute the configured search.

        A natural language word or phrase is the input.
        ---
        tag: search
        description: Search for a string
        requestBody:
            description: Search request
            required: true
            content:
                application/json:
                    schema:
                        $ref: '#/components/schemas/Search'
        responses:
            '200':
                description: Success
                content:
                    text/plain:
                        schema:
                            type: string
                            example: "Nominal search"
            '400':
                description: Malformed message
                content:
                    text/plain:
                        schema:
                            type: string

        """
        logger.debug(f"search:{json.dumps(request.json)}")
        response = {}
        try:
            app.logger.info (f"search: {json.dumps(request.json, indent=2)}")
            self.validate(request, component="Search")
            boosted = request.json.pop('boosted', False)

            api_request = None
            if boosted:
                api_request = dug().search_nboost(**request.json)
            else:
                api_request = dug().search_concepts(**request.json)

            response = self.create_response(
                result=api_request,
                message=f"Search result")
        except Exception as e:
            response = self.create_response(
                exception=e,
                message=f"Failed to execute search {json.dumps(request.json, indent=2)}.")
        return response


class DugSearchKGResource(DugResource):
    """ Execute a search """

    """ System initiation. """

    def post(self):
        """
        Execute the configured search.

        A natural language word or phrase is the input.
        ---
        tag: search_kg
        description: Search for a string across knowledge graphs
        requestBody:
            description: Search request
            required: true
            content:
                application/json:
                    schema:
                        $ref: '#/components/schemas/SearchKG'
        responses:
            '200':
                description: Success
                content:
                    text/plain:
                        schema:
                            type: string
                            example: "Nominal search"
            '400':
                description: Malformed message
                content:
                    text/plain:
                        schema:
                            type: string

        """
        logger.debug(f"search_kg:{json.dumps(request.json)}")
        response = {}
        try:
            app.logger.info(f"search_kg: {json.dumps(request.json, indent=2)}")
            self.validate(request, component="Search")
            response = self.create_response(
                result=dug().search_kg(**request.json),
                message=f"Search result")
        except Exception as e:
            response = self.create_response(
                exception=e,
                message=f"Failed to execute search {json.dumps(request.json, indent=2)}.")
        return response

class DugSearchVarResource(DugResource):
    """ Execute a search """

    """ System initiation. """

    def post(self):
        """
        Execute the configured search.

        A natural language word or phrase is the input.
        ---
        tag: search_kg
        description: Search for a string across knowledge graphs
        requestBody:
            description: Search request
            required: true
            content:
                application/json:
                    schema:
                        $ref: '#/components/schemas/SearchVar'
        responses:
            '200':
                description: Success
                content:
                    text/plain:
                        schema:
                            type: string
                            example: "Nominal search"
            '400':
                description: Malformed message
                content:
                    text/plain:
                        schema:
                            type: string

        """
        logger.debug(f"search_kg:{json.dumps(request.json)}")
        response = {}
        try:
            app.logger.info(f"search_var: {json.dumps(request.json, indent=2)}")
            self.validate(request, component="Search")
            response = self.create_response(
                result=dug().search_variables(**request.json),
                message=f"Search result")
        except Exception as e:
            response = self.create_response(
                exception=e,
                message=f"Failed to execute search {json.dumps(request.json, indent=2)}.")
        return response

""" Register endpoints. """
api.add_resource(DugSearchResource, '/search')
api.add_resource(DugSearchKGResource, '/search_kg')
api.add_resource(DugSearchVarResource, '/search_var')

if __name__ == "__main__":
   parser = argparse.ArgumentParser(description='Dug Search API')
   parser.add_argument('-p', '--port',  type=int, help='Port to run service on.', default=5551)
   parser.add_argument('-d', '--debug', help="Debug log level.", default=False, action='store_true')
   args = parser.parse_args ()

   """ Configure """
   if args.debug:
       debug = True
       logging.basicConfig(level=logging.DEBUG)
   logger.info (f"starting dug on port={args.port} with debug={args.debug}")
   app.run(host='0.0.0.0', port=args.port, debug=args.debug, threaded=True)



import json
class MissingNodeReferenceError(BaseException):
    pass

class MissingEdgeReferenceError(BaseException):
    pass

class QueryKG:
    def __init__(self, kg_json):
        self.kg = kg_json
        self.answers = kg_json.get('knowledge_map', [])
        self.question = kg_json["question_graph"]
        self.nodes = {node["id"]: node for node in kg_json.get('knowledge_graph', {}).get('nodes', [])}
        self.edges = {edge["id"]: edge for edge in kg_json.get('knowledge_graph', {}).get('edges', [])}

    def get_answer_subgraph(self, answer, include_node_keys=[], include_edge_keys=[]):

        # Get answer nodes
        answer_nodes = []
        for binding_id, binding_nodes in answer["node_bindings"].items():
            # Add node info for each node included in answer graph
            for answer_node in binding_nodes:
                # Throw error if node doesn't actually exist in 'nodes' section of knowledge graph
                if answer_node not in self.nodes:
                    err_msg = f"Unable to assemble subraph for answer:\n{json.dumps(answer, indent=2)}\n" \
                        f"Parent graph doesn't contain node info for: {answer_node}"
                    raise MissingNodeReferenceError(err_msg)

                # Add only node info that you actually want
                answer_nodes.append(self.get_node(answer_node, include_node_keys))

        # Get answer edges
        answer_edges = []
        for binding_id, binding_edges in answer["edge_bindings"].items():
            # Add edge info for each edge included in answer graph
            for answer_edge in binding_edges:
                # Throw error if edge doesn't actually exist in 'edges' section of knowledge graph
                if answer_edge not in self.edges:
                    err_msg = f"Unable to assemble subraph for answer:\n{json.dumps(answer, indent=2)}\n" \
                        f"Parent graph doesn't contain edge info for: {answer_edge}"
                    raise MissingEdgeReferenceError(err_msg)

                # Add only information from edge that you actually want
                answer_edges.append(self.get_edge(answer_edge, include_edge_keys))

        kg = {"knowledge_graph": {"nodes": answer_nodes,
                                  "edges": answer_edges},
              "knowledge_map": [answer],
              "question_graph": self.question}

        return QueryKG(kg)

    def get_node(self, node_id, include_node_keys=[]):
        # Return node with optionally subsetted information
        node = self.nodes[node_id]

        # Optionally subset to get only certain information columns
        if include_node_keys:
            node = {key: node[key] for key in include_node_keys}
        return node

    def get_edge(self, edge_id, include_edge_keys=[]):
        # Return edge with optionally subsetted information
        edge = self.edges[edge_id]

        # Optionally subset to include only certain info
        if include_edge_keys:
            edge = {key: edge[key] for key in include_edge_keys}
        return edge


class InvalidQueryError(BaseException):
    pass


class QueryFactory:

    # Class member list of valid data types that can be included in query
    data_types = ["phenotypic_feature", "gene", "disease", "chemical_substance",
                  "drug_exposure", "biological_process", "anatomical_entity"]

    # List of curie prefixes that are valid for certain curie types
    curie_map = {"disease": ["MONDO", "ORPHANET", "DOID"],
                 "phenotypic_feature": ["HP", "HPO", "EFO"],
                 "gene": ["HGNC", "NCBIGene"],
                 "chemical_substance": ["CHEBI", "PUBCHEM"],
                 "anatomical_entity": ["UBERON"]}

    def __init__(self, question_graph, source, curie_index=0):

        # List of terms that are going to be connected to make a query
        self.question_graph = question_graph

        # Index in question graph that will be matched against curies
        self.curie_index = curie_index

        # Query source (e.g. /schema or /graph/gamma/quick)
        self.source = source

        # Check to make sure curie index isn't out of range
        if self.curie_index >= len(self.question_graph):
            raise InvalidQueryError(f"Invalid query index ({curie_index})! Question graph only "
                                    f"contains {len(self.question_graph)} entries!")

        # Set the type of the curie for downstream curie checking
        self.curie_type = self.question_graph[self.curie_index]

        # Validate that all entries in question graph are actually valid types
        self.validate_factory()

    def validate_factory(self):
        # Check to make sure all the question types are valid
        for question in self.question_graph:
            if not question in QueryFactory.data_types:
                raise InvalidQueryError(f"Query contains invalid query type: {question}")

    def is_valid_curie(self, curie):
        # Return whether a curie can be used to create a valid query

        # Handle case where curie type has no limitations
        if self.curie_type not in QueryFactory.curie_map:
            return True

        # Otherwise only return true if current query contains one of the acceptable prefixes
        for curie_prefix in QueryFactory.curie_map[self.curie_type]:
            if curie.startswith(curie_prefix):
                return True

        # Curie doesn't start with an acceptable prefix
        return False

    def get_query(self, curie):

        # Return nothing if not valid curie
        if not self.is_valid_curie(curie):
            return None

        question = []
        seen = []
        curie_id = ""
        for i in range(len(self.question_graph)):
            query_type = self.question_graph[i]
            if self.question_graph.count(query_type) > 1:
                # Add alias to beginning of types we've seen before
                alias = f"{query_type[0:3]}{len([x for x in seen if x == query_type])}"
                query = f"{alias}:{query_type}"
            else:
                alias = query_type
                query = query_type

            # Set curie id to the alias if this is the correct index
            if i == self.curie_index:
                curie_id = alias

            # Append to list of query_types currently in query
            seen.append(query_type)
            # Append to list of actual terms that will appear in query
            question.append(query)

        # Build and return query
        return f"select {'->'.join(question)} from '{self.source}' where {curie_id}='{curie}'"



from elasticsearch import Elasticsearch
from dug.annotate import TOPMedStudyAnnotator
from dug.utils import BioLinkPURLerizer
import dug.tranql as tql
import argparse
import logging
import glob
import json
import requests
import traceback
import os

logger = logging.getLogger (__name__)

logging.getLogger("elasticsearch").setLevel(logging.WARNING)

class SearchException (Exception):
    def __init__(self, message, details):
        self.message = message
        self.details = details

class Search:
    """ Search -
    1. Lexical fuzziness; (a) misspellings - a function of elastic.
    2. Fuzzy ontologically;
       (a) expand based on core queries
         * phenotype->study
         * phenotype->disease->study
         * disease->study
         * disease->phenotype->study
    """
    def __init__(self, host=os.environ.get('ELASTIC_API_HOST'), port=9200, indices=['concepts_index', 'variables_index', 'kg_index']):
        logger.debug (f"Connecting to elasticsearch host: {host} at port: {port}")
        self.indices = indices
        self.crawlspace = "crawl"
        self.host = os.environ.get ('ELASTIC_API_HOST', 'localhost')
        self.username = os.environ.get ('ELASTIC_USERNAME', 'elastic')
        self.password = os.environ.get ('ELASTIC_PASSWORD', 'changeme')
        self.nboost_host = os.environ.get('NBOOST_API_HOST', 'nboost')
        self.hosts = [
            {
                'host' : self.host,
                'port' : port
            }
        ]
        logger.debug (f"Authenticating as user {self.username} to host:{self.hosts}")
        self.es = Elasticsearch (hosts=self.hosts,
                                 http_auth=(self.username, self.password))

        if self.es.ping():
            logger.info ('connected to elasticsearch')
            self.init_indices ()
        else:
            print (f"Unable to connect to elasticsearch at {host}:{port}")
            logger.error (f"Unable to connect to elasticsearch at {host}:{port}")
            raise SearchException (
                message='failed to connect to elasticsearch',
                details=f"connecting to host {host} and port {port}")

    def clean (self):
        self.es.indices.delete ("*")

    def init_indices (self):
        settings = {}
        
        # kg_index
        settings['kg_index'] = {
            "settings": {
                "number_of_shards": 1,
                "number_of_replicas": 0
            },
            "mappings": {
                "properties": {
                    "name": {
                        "type": "text"
                    },
                    "type": {
                        "type": "text"
                    }
                }
            }
        }

        # concepts_index
        settings['concepts_index'] = {
            "settings": {
                "index.mapping.coerce": "false",
                "number_of_shards": 1,
                "number_of_replicas": 0
            }, 
            "mappings": {
                "dynamic": "strict",
                "properties": {
                    "id": {"type": "text", "fields": {"keyword": {"type": "keyword"}}},
                    "name": {"type": "text"},
                    "description": {"type": "text"},
                    "type": {"type": "keyword"},
                    "search_terms": {"type": "text"},
                    "identifiers": {
                        "properties": {
                            "id": {"type": "text", "fields": {"keyword": {"type": "keyword"}}},
                            "label": {"type": "text"},
                            "equivalent_identifiers": {"type": "keyword"},
                            "type": {"type": "keyword"},
                            "synonyms": {"type": "text"}
                            }
                    },
                    "optional_terms": {"type": "text"},
                    "concept_action": {"type": "text"}
                }
            }
        }

        # variables_index
        settings['variables_index'] = {
            "settings": {
                "index.mapping.coerce": "false",
                "number_of_shards": 1,
                "number_of_replicas": 0
            }, 
            "mappings": {
                "dynamic": "strict",
                "properties": {
                    "element_id": {"type": "text", "fields": {"keyword": {"type": "keyword"}}},
                    "element_name": {"type": "text"},
                    "element_desc": {"type": "text"},
                    "element_action": {"type": "text"},
                    "search_terms": {"type": "text"},
                    "identifiers": {"type": "keyword"},
                    "collection_id": {"type": "text", "fields": {"keyword": {"type": "keyword"}}},
                    "collection_name": {"type": "text"},
                    "collection_desc": {"type": "text"},
                    "collection_action": {"type": "text"},
                    "data_type": {"type": "text"},
                }
            }
        }

        logger.info (f"creating indices: {self.indices}")
        for index in self.indices:
            try:
                if self.es.indices.exists(index=index):
                    logger.info (f"Ignoring index {index} which already exists.")
                else:
                    result = self.es.indices.create (
                        index=index,
                        body=settings[index],
                        ignore=400)
                    logger.info (f"result created index {index}: {result}")
            except Exception as e:
                logger.error (f"exception: {e}")
                raise e

    def index_doc (self, index, doc, doc_id):
        self.es.index (
            index=index,
            id=doc_id,
            body=doc)
    
    def update_doc (self, index, doc, doc_id):
        self.es.update (
            index=index,
            id=doc_id,
            body=doc
        )

    def search (self, index, query, offset=0, size=None, fuzziness=1):
        """
        Query type is now 'query_string'.
        query searches multiple fields
        if search terms are surrounded in quotes, looks for exact matches in any of the fields
        AND/OR operators are natively supported by elasticesarch queries
        """
        query = {
            'query_string': {
                'query' : query,
                'fuzziness' : fuzziness,
                'fields': ['name', 'description', 'instructions', 'search_targets', 'optional_targets'],
                'quote_field_suffix': ".exact"
            }

        }
        body = json.dumps({'query': query})
        total_items = self.es.count(body=body, index=index)
        search_results = self.es.search(
            index=index,
            body=body,
            filter_path=['hits.hits._id', 'hits.hits._type', 'hits.hits._source'],
            from_=offset,
            size=size
        )
        search_results.update({'total_items': total_items['count']})
        return search_results

    def search_concepts (self, index, query, offset=0, size=None, fuzziness=1, prefix_length=3):
        """
        Changed to query_string for and/or and exact matches with quotations.
        """
        query = {
            'query_string': {
                'query' : query,
                'fuzziness' : fuzziness,
                'fuzzy_prefix_length': prefix_length,
                'fields': ["name", "description", "search_terms", "optional_terms"],
                'quote_field_suffix': ".exact"
            },
        }
        body = json.dumps({'query': query})
        total_items = self.es.count(body=body, index=index)
        search_results = self.es.search(
            index=index,
            body=body,
            filter_path=['hits.hits._id', 'hits.hits._type', 'hits.hits._source'],
            from_=offset,
            size=size
        )
        search_results.update({'total_items': total_items['count']})
        return search_results

    def search_variables(self, index, concept, query, offset=0, size=None, fuzziness=1, prefix_length=3):
        """
        In variable seach, the concept MUST match one of the indentifiers in the list
        The query can match search_terms (hence, "should") for ranking.
        """
        query = {
            'bool': {
                'must': {
                        "match": {
                            "identifiers": concept
                        }
                    },
                'should': {
                    'query_string': {
                        "query": query,
                        "fuzziness": fuzziness,
                        "fuzzy_prefix_length": prefix_length,
                        "default_field": "search_terms"
                    }
                }
            }
        }
        body = json.dumps({'query': query})
        total_items = self.es.count(body=body, index=index)
        search_results = self.es.search(
            index=index,
            body=body,
            filter_path=['hits.hits._id', 'hits.hits._type', 'hits.hits._source'],
            from_=offset,
            size=size
        )
        search_results.update({'total_items': total_items['count']})
        return search_results

    def search_kg(self, index, unique_id, query, offset=0, size=None, fuzziness=1, prefix_length=3):
        """
        In knowledge graph search seach, the concept MUST match the unique ID
        The query MUST match search_targets.  The updated query allows for
        fuzzy matching and for the default OR behavior for the query.
        """
        query = {
            "bool": {
                "must": [
                    {"term": {
                        "concept_id.keyword": unique_id
                        }
                    },
                    {'query_string': {
                        "query": query,
                        "fuzziness": fuzziness,
                        "fuzzy_prefix_length": prefix_length,
                        "default_field": "search_targets"
		                }
                    }
                ]
            }
        }
        body = json.dumps({'query': query})
        total_items = self.es.count(body=body, index=index)
        search_results = self.es.search(
            index=index,
            body=body,
            filter_path=['hits.hits._id', 'hits.hits._type', 'hits.hits._source'],
            from_=offset,
            size=size
        )
        search_results.update({'total_items': total_items['count']})
        return search_results

    def search_nboost(self, index, query, offset=0, size=10, fuzziness=1):
        """
        Query type is now 'query_string'.
        query searches multiple fields
        if search terms are surrounded in quotes, looks for exact matches in any of the fields
        AND/OR operators are natively supported by elasticesarch queries
        """
        nboost_query = {
            'nboost': {
                'uhost': f"{self.username}:{self.password}@{self.host}",
                'uport': self.hosts[0]['port'],
                'cvalues_path': '_source.description',
                'query_path': 'body.query.query_string.query',
                'size': size,
                'from': offset,
                'default_topk': size
            },
            'query': {
                'query_string': {
                    'query': query,
                    'fuzziness': fuzziness,
                    'prefix_length': prefix_length,
                    'fields': ['name', 'description', 'instructions', 'search_targets', 'optional_targets'],
                    'quote_field_suffix': ".exact"
                }
            }
        }

        return requests.post(url=f"http://{self.nboost_host}:8000/{index}/_search", json=nboost_query).json()

    def make_crawlspace (self):
        if not os.path.exists (self.crawlspace):
            try:
                os.makedirs (self.crawlspace)
            except Exception as e:
                print (f"-----------> {e}")
                traceback.print_exc ()

    def crawl (self, tags, variables, index, queries, min_score=0.2,
               include_node_keys=["id", "name", "synonyms"], include_edge_keys=[],
               query_exclude_identifiers=[]):
        tranql_endpoint = "https://tranql.renci.org/tranql/query?dynamic_id_resolution=true&asynchronous=false"
        headers = {
            "accept" : "application/json",
            "Content-Type" : "text/plain"
        }

        self.make_crawlspace ()
        for tag in tags:
            # Get subset of variables for this tag
            tagged_variables = [variable for variable in variables if int(variable["tag_pk"]) == int(tag["pk"])]
            tag_indexed = False

            logging.debug(f"Doing variables with tag: {tag['title']}")
            for identifier in tag["identifiers"]:

                logging.debug(f"Doing id: {identifier}")
                ''' Resolve the phenotype to identifiers. '''

                # Boolean switch for whether a knowledge graph has been returned for the current identifier
                # If no knowledge graphs are returned by TranQL just put a normal record with no nodes/KG
                identifier_indexed = False

                # Loop through each query and try to add the answers to the search index
                for query_name, query_factory in queries.items():

                    # Skip identifiers that didn't normalize
                    if not tag["identifiers"][identifier]["label"]:
                        logging.debug(f"Skipping non-normalized identifier: {identifier}")
                        continue

                    # Skip identifier if it's in the exclude list
                    if identifier in query_exclude_identifiers:
                        logging.debug(f"Skipping TranQL query for exclude listed identifier: {identifier}")
                        continue

                    # Skip query if a file exists in the crawlspace exists already
                    filename = f"{self.crawlspace}/{identifier}_{query_name}.json"
                    if os.path.exists(filename):
                        logger.info(f"identifier {identifier} is already crawled.")
                        continue

                    # Skip query if the identifier is not a valid query for the query class
                    if not query_factory.is_valid_curie(identifier):
                        logger.info(f"identifier {identifier} is not valid for query type {query_name}. Skipping!")
                        continue

                    # Submit query to TranQL
                    query = query_factory.get_query(identifier)
                    logger.info (query)
                    response = requests.post(
                        url = tranql_endpoint,
                        headers = headers,
                        data = query).json ()

                    print(response)

                    # Case: Skip if empty KG
                    if not len(response["knowledge_graph"]["nodes"]):
                        logging.debug(f"Did not find a knowledge graph for {query}")
                        continue

                    # Dump out to file if there's a knowledge graph
                    with open(filename, 'w') as stream:
                        json.dump(response, stream, indent=2)

                    # Get nodes in knowledge graph hashed by ids for easy lookup
                    kg = tql.QueryKG(response)

                    # Create ES entries for each variable for each answer
                    for answer in kg.answers:

                        # Filter out answers that fall below a minimum score
                        # TEMPORARY: Robokop stopped including scores temporarily so ignore these for time being
                        # We don't know how this filtering works; let's bring back everything for now, so we keep all synonyms
                        #if "score" in answer and answer["score"] < min_score:
                        #    continue
                        logger.debug(f"Answer: {answer}")

                        # Get subgraph containing only information for this answer
                        try:
                            # Temporarily surround in try/except because sometimes the answer graphs
                            # contain invalid references to edges/nodes
                            # This will be fixed in Robokop but for now just silently warn if answer is invalid
                            answer_kg = kg.get_answer_subgraph(answer,
                                                               include_node_keys=include_node_keys,
                                                               include_edge_keys=include_edge_keys)

                            # Get list of nodes for making a unique ID for elastic search
                            answer_node_ids = list(answer_kg.nodes.keys())

                        except tql.MissingNodeReferenceError:
                            # TEMPORARY: Skip answers that have invalid node references
                            # Need this to be fixed in Robokop
                            logger.warning("Skipping answer due to presence of non-preferred id! "
                                           "See err msg for details.")
                            continue
                        except tql.MissingEdgeReferenceError:
                            # TEMPORARY: Skip answers that have invalid edge references
                            # Need this to be fixed in Robokop
                            logger.warning("Skipping answer due to presence of invalid edge reference! "
                                           "See err msg for details.")
                            continue

                        # Add each variable to ES with info specific to current answer
                        self.index_tagged_variables(tag,
                                                    tagged_variables,
                                                    index,
                                                    identifier=identifier,
                                                    knowledge_graph=answer_kg.kg,
                                                    query_name=query_name,
                                                    answer_node_ids=answer_node_ids)

                        # Set boolean flag that at least one answer has been added to elastic for an identifier
                        # Now we know we don't need to write a dummy record for this identifier
                        identifier_indexed = True
                        tag_indexed = True

                # Write textual entries for identifiers that didn't return KG from TranQL
                # Ensures we don't lose the ability to search on tag identifier labels returned from Monarch
                if not identifier_indexed and tag["identifiers"][identifier]["label"]:
                    self.index_tagged_variables(tag,
                                                tagged_variables,
                                                index,
                                                identifier=identifier)

                    # Indicate that tag has been indexed at least once
                    # Now we know we don't need to write a dummy record for this tag
                    tag_indexed = True

            # Handle the exceptional case where a tag doesn't actually have any identifiers
            # (i.e. Monarch didn't know what it was)
            # We need to add records so we can at least still search on the tag's generic name/description
            if not tag_indexed:
                self.index_tagged_variables(tag,
                                            tagged_variables,
                                            index)
                        
    def index_tagged_variables(self, tag, variables, index, identifier="", knowledge_graph={}, query_name="", answer_node_ids=[]):
        # Internal class helper method for writing a list of tagged variables to Elastic Search

        # Use identifier label as name if identifier exists
        # Some tags may not have identifiers (e.g. when Monarch fails to return something) so just use empty string
        name = tag["identifiers"][identifier]["label"] if identifier else ""

        # Get all the stuff you want to make searchable out of the knowledge graph into one flat array
        kg_search_targets = []
        for node in knowledge_graph.get("knowledge_graph", {}).get("nodes", []):
            kg_search_targets.append(node["name"])
            kg_search_targets += node["synonyms"]
        
        # Add synonyms if no answers from TranQL and tag has identifier
        if not len(kg_search_targets) and len(name):
            kg_search_targets = tag['identifiers'][identifier]['synonyms']

        # TODO: Add synonyms here
        for variable in variables:
            doc = {"name": name,
                   "id": identifier,
                   "var": variable["variable_id"].replace("TOPMED.VAR:", ""),
                   "tag": variable["tag_pk"],
                   "description": tag["description"],
                   "instructions": tag["instructions"],
                   "study": variable["study_id"].replace("TOPMED.STUDY:", ""),
                   "study_name": variable["study_name"],
                   "knowledge_graph": knowledge_graph,
                   "search_targets": kg_search_targets}

            # Create unique ID
            if answer_node_ids and query_name:
                # Case: Variable is created from KG query and needs query and answer nodes to be unique
                logger.debug("Indexing TranQL query answer...")
                unique_doc_id = f"{doc['id']}_{doc['study']}_{doc['var']}_{'_'.join(answer_node_ids)}_{query_name}"
            elif doc['id']:
                logger.debug("Indexing identifier that didn't return anything from TranQL")
                # Case: Variable is created from one of a tag's identifiers and identifier can be used as unique
                unique_doc_id = f"{doc['id']}_{doc['study']}_{doc['var']}"
            else:
                # Case: Variable doesn't have any identifiers (monarch failed)
                # The study name/variable name will be fine for unique
                logger.debug("Indexing generic tagged variable...")
                unique_doc_id = f"{doc['study']}_{doc['var']}"

            logger.debug(f"ElasticSearch ID: {unique_doc_id}\n{json.dumps(doc, indent=2)}")

            """ Index the document. """
            self.index_doc(
                index=index,
                doc=doc,
                doc_id=unique_doc_id)
    
    def crawl_by_tag (self, tags, variables, tag_index, kg_index, queries, min_score=0.2,
                      include_node_keys=["id", "name", "synonyms"], include_edge_keys=[],
                      query_exclude_identifiers=[]):

        '''
        This version of tagged crawl starts from preferred identifiers linked to tags
        (as biological_entity type) and performs all available queries, stashing knowledge graphs
        at the tag level
        '''
        tranql_endpoint = "https://tranql.renci.org/tranql/query?dynamic_id_resolution=true&asynchronous=false"
        headers = {
            "accept" : "application/json",
            "Content-Type" : "text/plain"
        }

        self.make_crawlspace ()
        for tag in tags:
            ## TOPMed Studies, Variables
            studies = {}
            
            ## Search Targets
            tag['search_targets'] = [] # Initialize Search Target section
            tag['optional_targets'] = [] # Initialize answer targets
            
            ## Set tag_indexed to False; case where tag does not return KGs from TranQL
            tag_indexed = False

            tagged_variables = [variable for variable in variables if int(variable["tag_pk"]) == int(tag["pk"])]
            for variable in tagged_variables:
                study_id = variable['study_id'].replace("TOPMED.STUDY:", "")
                study_name = variable['study_name']
                variable_id = variable['variable_id'].replace("TOPMED.VAR:", "")
                
                if study_id not in studies:  
                    studies[study_id] = {
                        "study_id": study_id,
                        "study_name": study_name,
                        "variables": [variable_id]
                    }
                else:
                    studies[study_id]['variables'].append(variable_id)

            # Convert Studies to a List
            tag['studies'] = list(studies.values())
            
            ## Identifiers
            # Deal with identifiers to reduce indexing size
            identifier_list = []
            for identifier in tag['identifiers']:
                # Deal with Identifiers
                identifier_dict = tag['identifiers'][identifier]
                identifier_dict['identifier'] = identifier
                identifier_list.append(identifier_dict)

                # Add identifier Search Targets
                if tag["identifiers"][identifier]["label"]:
                    tag['search_targets'] += tag['identifiers'][identifier].get('synonyms',[])
                    tag['search_targets'].append(tag['identifiers'][identifier]['label'])

            tag['identifier_list'] = identifier_list

            ## Queries
            for identifier in tag["identifiers"]:
                logging.debug(f"Doing id: {identifier}")
                ''' Resolve the phenotype to identifiers. '''
                
                # skip identifiers that don't normalize, or are excluded
                if not tag["identifiers"][identifier]["label"]:
                    logging.debug(f"Skipping non-normalized identifier: {identifier}")
                    continue
                
                if identifier in query_exclude_identifiers:
                    logging.debug(f"Skipping TranQL query for exclude listed identifier: {identifier}")
                    continue

                for query_name, query_factory in queries.items():
                    
                    # Skip query if the identifier is not a valid query for the query class
                    if not query_factory.is_valid_curie(identifier):
                        logger.info(f"identifier {identifier} is not valid for query type {query_name}. Skipping!")
                        continue
                    
                    filename = f"{self.crawlspace}/{identifier}_{query_name}.json"
                    # Skip TranQL query if a file exists in the crawlspace exists already, but continue w/ answers
                    if os.path.exists(filename):
                        logger.info(f"identifier {identifier} is already crawled. Skipping TranQL query.")
                        with open (filename, 'r') as stream:
                            response = json.load(stream)

                    else:
                        query = query_factory.get_query(identifier)
                        logger.info (query)
                        response = requests.post(
                            url = tranql_endpoint,
                            headers = headers,
                            data = query).json ()

                        # Case: Skip if empty KG 
                        if not len(response['knowledge_graph']['nodes']):
                            logging.debug(f"Did not find a knowledge graph for {query}")
                            continue # continue out of loop
                        
                        # Dump out to file if there's a knowledge graph
                        with open(filename, 'w') as stream:
                            json.dump(response, stream, indent=2)    
                    
                    # Get nodes in knowledge graph hashed by ids for easy lookup
                    kg = tql.QueryKG(response)

                    for answer in kg.answers:

                        # Filter out answers that fall below a minimum score
                        # TEMPORARY: Robokop stopped including scores temporarily so ignore these for time being
                        # We don't know how this filtering works; let's bring back everything for now, so we keep all synonyms
                        #if "score" in answer and answer["score"] < min_score:
                        #    continue
                        logger.debug(f"Answer: {answer}")

                        # Get subgraph containing only information for this answer
                        try:
                            # Temporarily surround in try/except because sometimes the answer graphs
                            # contain invalid references to edges/nodes
                            # This will be fixed in Robokop but for now just silently warn if answer is invalid
                            answer_kg = kg.get_answer_subgraph(answer,
                                                               include_node_keys=include_node_keys,
                                                               include_edge_keys=include_edge_keys)

                            # Get list of nodes for making a unique ID for elastic search
                            answer_node_ids = list(answer_kg.nodes.keys())

                        except tql.MissingNodeReferenceError:
                            # TEMPORARY: Skip answers that have invalid node references
                            # Need this to be fixed in Robokop
                            logger.warning("Skipping answer due to presence of non-preferred id! "
                                           "See err msg for details.")
                            continue
                        except tql.MissingEdgeReferenceError:
                            # TEMPORARY: Skip answers that have invalid edge references
                            # Need this to be fixed in Robokop
                            logger.warning("Skipping answer due to presence of invalid edge reference! "
                                           "See err msg for details.")
                            continue

                        # Add answer synonyms to tag's list of optional targets
                        for node in answer_kg.kg.get('knowledge_graph', {}).get('nodes', []):
                            tag['optional_targets'].append(node['name'])
                            tag['optional_targets'] += node['synonyms']

                        # Add answer to knowledge graph ES index
                        self.index_kg_answer(tag,
                                             kg_index,
                                             curie_id=identifier,
                                             knowledge_graph=answer_kg.kg,
                                             query_name=query_name,
                                             answer_node_ids=answer_node_ids)

            # Add tag with all info to tag index
            self.index_tag(tag, tag_index)

    def crawl_by_concept (self, concepts, concept_index, kg_index, queries, min_score=0.2,
                        include_node_keys=["id", "name", "synonyms"], include_edge_keys=[],
                        query_exclude_identifiers=[]):

        '''
        This version of tagged crawl starts from identifiers within concepts.
        If an ontological term, the concept will only have one ontology identifier: itself
        If it is, for instance, a TOPMed phenotype concept, there will be multiple ontology identifiers.
        '''
        tranql_endpoint = "https://tranql.renci.org/tranql/query?dynamic_id_resolution=true&asynchronous=false"
        headers = {
            "accept" : "application/json",
            "Content-Type" : "text/plain"
        }

        self.make_crawlspace ()
        for concept in concepts:
            concepts[concept]['optional_terms'] = [] # Initialize answer targets
            
            ## Set tag_indexed to False; case where tag does not return KGs from TranQL
            concept_indexed = False

            ## Queries
            for identifier in concepts[concept]["identifiers"]:
                logging.debug(f"Doing id: {identifier}")
                ''' Resolve the phenotype to identifiers. '''
                
                # skip identifiers that don't normalize, or are excluded
                if not concepts[concept]["identifiers"][identifier]["label"]:
                    logging.debug(f"Skipping non-normalized identifier: {identifier}")
                    continue
                
                if identifier in query_exclude_identifiers:
                    logging.debug(f"Skipping TranQL query for exclude listed identifier: {identifier}")
                    continue

                for query_name, query_factory in queries.items():
                    
                    # Skip query if the identifier is not a valid query for the query class
                    if not query_factory.is_valid_curie(identifier):
                        logger.info(f"identifier {identifier} is not valid for query type {query_name}. Skipping!")
                        continue
                    
                    filename = f"{self.crawlspace}/{identifier}_{query_name}.json"
                    # Skip TranQL query if a file exists in the crawlspace exists already, but continue w/ answers
                    if os.path.exists(filename):
                        logger.info(f"identifier {identifier} is already crawled. Skipping TranQL query.")
                        with open (filename, 'r') as stream:
                            response = json.load(stream)

                    else:
                        query = query_factory.get_query(identifier)
                        logger.info (query)
                        response = requests.post(
                            url = tranql_endpoint,
                            headers = headers,
                            data = query).json ()

                        # Case: Skip if empty KG 
                        if not len(response['knowledge_graph']['nodes']):
                            logging.debug(f"Did not find a knowledge graph for {query}")
                            continue # continue out of loop
                        
                        # Dump out to file if there's a knowledge graph
                        with open(filename, 'w') as stream:
                            json.dump(response, stream, indent=2)    
                    
                    # Get nodes in knowledge graph hashed by ids for easy lookup
                    kg = tql.QueryKG(response)

                    for answer in kg.answers:

                        # Filter out answers that fall below a minimum score
                        # TEMPORARY: Robokop stopped including scores temporarily so ignore these for time being
                        # We don't know how this filtering works; let's bring back everything for now, so we keep all synonyms
                        #if "score" in answer and answer["score"] < min_score:
                        #    continue
                        logger.debug(f"Answer: {answer}")

                        # Get subgraph containing only information for this answer
                        try:
                            # Temporarily surround in try/except because sometimes the answer graphs
                            # contain invalid references to edges/nodes
                            # This will be fixed in Robokop but for now just silently warn if answer is invalid
                            answer_kg = kg.get_answer_subgraph(answer,
                                                            include_node_keys=include_node_keys,
                                                            include_edge_keys=include_edge_keys)

                            # Get list of nodes for making a unique ID for elastic search
                            answer_node_ids = list(answer_kg.nodes.keys())

                        except tql.MissingNodeReferenceError:
                            # TEMPORARY: Skip answers that have invalid node references
                            # Need this to be fixed in Robokop
                            logger.warning("Skipping answer due to presence of non-preferred id! "
                                        "See err msg for details.")
                            continue
                        except tql.MissingEdgeReferenceError:
                            # TEMPORARY: Skip answers that have invalid edge references
                            # Need this to be fixed in Robokop
                            logger.warning("Skipping answer due to presence of invalid edge reference! "
                                        "See err msg for details.")
                            continue

                        # Add answer synonyms to tag's list of optional targets
                        for node in answer_kg.kg.get('knowledge_graph', {}).get('nodes', []):
                            concepts[concept]['optional_terms'].append(node['name'])
                            concepts[concept]['optional_terms'] += node['synonyms']

                        # Add answer to knowledge graph ES index
                        self.index_kg_answer(concepts[concept],
                                            kg_index,
                                            curie_id=identifier,
                                            knowledge_graph=answer_kg.kg,
                                            query_name=query_name,
                                            answer_node_ids=answer_node_ids)

            # Add tag with all info to tag index
            self.index_concept(concepts[concept], concepts_index)


    def index_tag(self, tag, index):
        # Create the Doc
        doc = {
            'tag_id': tag['id'],
            'pk': tag['pk'],
            'name': tag['title'],
            'description': tag['description'],
            'instructions': tag['instructions'],
            'search_targets': list(set(tag['search_targets'])),  # Make unique if duplicates
            'optional_targets': list(set(tag['optional_targets'])),
            'studies': tag['studies'],
            'identifiers': tag['identifier_list']
        }
        # DEBUG: For writing elasticsearch documents to JSON
        #with open(f'new_doc_structure/{tag["id"]}.json', 'w') as stream:
        #    json.dump(doc, stream, indent=2)

        """ Index the document. """
        self.index_doc(
            index=index,
            doc=doc,
            doc_id=doc['tag_id'])

    def index_concept(self, concept, index):
        """ Index the document. """
        identifier_dict = concept['identifiers']
        concept['identifiers'] = []
        # Rearrange identifiers
        for identifier in identifier_dict:
            r_identifier = identifier_dict[identifier]
            r_identifier['id'] = identifier
            concept['identifiers'].append(r_identifier)
        
        # Make optional_terms unique
        concept['optional_terms'] = list(set(concept['optional_terms'])) 
        self.index_doc(
            index=index,
            doc=concept,
            doc_id=concept['id'])
    
    def index_variables(self, variables, index):
        for variable in variables:
            if not self.es.exists(index,variable['element_id']):
                self.index_doc(
                    index=index,
                    doc=variable,
                    doc_id=variable['element_id'])
            else:
                results = self.es.get(index, variable['element_id'])
                identifiers = results['_source']['identifiers'] + variable['identifiers'] 
                doc = {"doc" :{}}
                doc['doc']['identifiers'] = identifiers
                self.update_doc(index = index, doc = doc, doc_id = variable['element_id'])

                

    def index_kg_answer(self, concept, index, curie_id, knowledge_graph, query_name, answer_node_ids):
        answer_synonyms = []
        for node in knowledge_graph.get('knowledge_graph', {}).get('nodes', []):
            # Don't add curie synonyms to knowledge graph
            #  We only want to return KG answer if it relates to user query
            if node["id"] == curie_id:
                continue
            answer_synonyms.append(node['name'])
            answer_synonyms += node['synonyms']

        # Create the Doc
        doc = {
            'concept_id': concept['id'],
            'search_targets': answer_synonyms,  # Make unique if duplicates
            'knowledge_graph': knowledge_graph
        }
        # Create unique ID
        logger.debug("Indexing TranQL query answer...")
        unique_doc_id = f"{doc['concept_id']}_{'_'.join(answer_node_ids)}_{query_name}"

        # DEBUG: For writing elasticsearch documents to JSON
        #with open(f'new_doc_structure/{unique_doc_id}.json', 'w') as stream:
        #    json.dump(doc, stream, indent=2)

        """ Index the document. """
        self.index_doc(
            index=index,
            doc=doc,
            doc_id=unique_doc_id)
  
    def index (self, index):
        self.make_crawlspace ()
        files = glob.glob (f"{self.crawlspace}/*.json")
        for f in files:
            print (f"index {f}")
            with open (f, 'r') as stream:
                kg = None
                try:
                    kg = json.load (stream)
                except Exception as e:
                    logger.error (f"Failed reading: {f}")
                    logger.error (e)
                if not kg:
                    continue
                #print ([ x for x in kg ])
                nodes = kg.get('knowledge_graph',{}).get('nodes',[])
                print (f"kg: {nodes}")
                doc = {
                    'diseases' : []
                }

                root_id = None
                for node in nodes:
                    identifier = node['id']
                    if identifier.startswith ('HP'):
                        doc['id'] = identifier
                        root_id = identifier
                        doc['name'] = node['synonyms']
                    else:
                        doc['diseases'].append (identifier)

                print (f"{json.dumps(doc, indent=2)}")
                search.index_doc (
                    index=index,
                    doc=doc,
                    doc_id=root_id)

if __name__ == '__main__':

    db_url_default = "http://" + os.environ.get('NEO4J_HOST', 'localhost') + ":" + os.environ.get('NEO4J_PORT',
                                                                                                  '7474') + "/db/data"
    
    parser = argparse.ArgumentParser(description='TranQL-Search')
    parser.add_argument('--clean', help="Clean", default=False, action='store_true')
    parser.add_argument('--index', help="Index", default=False, action='store_true')

    # Add mutually exclusive group for whether to do normal crawl or whether to crawl by tags
    crawl_group = parser.add_mutually_exclusive_group()
    crawl_group.add_argument('--crawl',
                             help="Crawl tagged variables and add to elastic search index",
                             default=False,
                             action='store_true')
    crawl_group.add_argument('--crawl-by-tag',
                             help="Crawl tagged variables and organize elastic search by tag",
                             default=False,
                             action='store_true',
                             dest="crawl_by_tag")
    crawl_group.add_argument('--crawl-by-concept',
                             help="Crawl tagged variables and organize elastic search by concept",
                             default=False,
                             action='store_true',
                             dest="crawl_by_concept")

    # Add mutually exclusive group for whether crawl inputs are coming from file or TranQL
    crawl_input_group = parser.add_mutually_exclusive_group()
    crawl_input_group.add_argument('--crawl-file',
                                   help='Input file containing things you want to crawl/index',
                                   dest="crawl_file")

    # Minimum score for Robokop answer to be included in ElasticSearch Index
    parser.add_argument('--min-tranql-score', help='Minimum score to consider an answer from TranQL',
                        dest="min_tranql_score",
                        default=0.2, type=float)

    parser.add_argument('--index_p1', help="Index - Phase 1 - local graph database rather than Translator query.",
                        default=False, action='store_true')

    parser.add_argument('--query', help="Query", action="store", dest="query")
    parser.add_argument('--query-kg', help="Query Knowledge graph", action="store", dest="query_kg")
    parser.add_argument('--kg-id', help="ID of Knowledge graph to query", action="store", dest="query_kg_id")
    parser.add_argument('--query-variables', help="Query Variables Index", action="store", dest="query_variables")
    parser.add_argument('--concept-id', help="ID of concept to query", action = "store", dest = "query_concept_id")

    parser.add_argument('--elastic-host', help="Elasticsearch host", action="store", dest="elasticsearch_host",
                        default=os.environ.get('ELASTIC_API_HOST', 'localhost'))
    parser.add_argument('--elastic-port', help="Elasticsearch port", action="store", dest="elasticsearch_port",
                        default=os.environ.get('ELASTIC_API_PORT', 9200))

    parser.add_argument('--db-url', help='database url', default=db_url_default)
    parser.add_argument('--db-username', help='database username', default='neo4j')
    parser.add_argument('--db-password', help='database password', default=os.environ['NEO4J_PASSWORD'])
    args = parser.parse_args ()

    logging.basicConfig(level=logging.DEBUG)
    concepts_index = "concepts_index"
    variables_index = "variables_index"
    kg_index = "kg_index"
    search = Search (host=args.elasticsearch_host,
                     port=args.elasticsearch_port,
                     indices=[concepts_index, variables_index, kg_index])

    if args.clean:
        search.clean ()
    elif args.query:
        val = search.search (index=concepts_index, query=args.query)
        if 'hits' in val:
            for hit in val['hits']['hits']:
                print (hit)
                print (f"{hit['_source']}")
    elif args.query_kg:
        # Throw error if user didn't provide a knowledge graph tag
        if not args.query_kg_id:
            parser.error("Must include knowledge graph id (--kg-id <id>) when querying knowledge graph!")
        print(args.query_kg)
        print(args.query_kg_id)
        val = search.search_kg(index=kg_index, unique_id=args.query_kg_id, query=args.query_kg)
        if 'hits' in val:
            for hit in val['hits']['hits']:
                print(hit)
                print(f"{hit['_source']}")
    elif args.query_variables:
        # Throw error if user didn't provide a knowledge graph tag
        if not args.query_concept_id:
            parser.error("Must include concept id (--concept-id <id>) when querying variables!")
        print(args.query_variables)
        print(args.query_concept_id)
        val = search.search_variables(index=variables_index, concept=args.query_concept_id, query=args.query_variables)
        if 'hits' in val:
            for hit in val['hits']['hits']:
                print(hit)
                print(f"{hit['_source']}")
    
    elif args.crawl or args.crawl_by_tag or args.crawl_by_concept:
        # Throw error if user hasn't specified where crawl arguments are coming from
        if not args.crawl_file:
            parser.error("Crawl must specify whether to get inputs from file "
                         "(--crawl-file <file>) or TranQL (--crawl-tranql)")

        config = {
            'annotator': "https://api.monarchinitiative.org/api/nlp/annotate/entities?min_length=4&longest_only=false&include_abbreviation=false&include_acronym=false&include_numbers=false&content=",
            'normalizer': "https://nodenormalization-sri.renci.org/get_normalized_nodes?curie=",
            'synonym_service': "https://onto.renci.org/synonyms/",
            'ontology_metadata': "https://api.monarchinitiative.org/api/bioentity/",
            'password': os.environ['NEO4J_PASSWORD'],
            'username': 'neo4j',
            'db_url': db_url_default,
            'redis_host': os.environ.get('REDIS_HOST', 'localhost'),
            'redis_port': os.environ.get('REDIS_PORT', 6379),
            'redis_password': os.environ.get('REDIS_PASSWORD', ''),
        }

        config['username'] = args.db_username
        config['password'] = args.db_password
        config['db_url'] = args.db_url

        # Create annotator object
        annotator = TOPMedStudyAnnotator(config=config)

        # Use file extension to determine how to parse for now
        # Eventually we'll need something more sophisticated as we get more types
        if args.crawl_file.endswith(".csv"):
            # Read in pre-harmonized variables (tagged variables we're calling them) from csv
            variables, tags = annotator.load_tagged_variables(args.crawl_file)
            concepts = annotator.annotate(tags)
        elif args.crawl_file.endswith(".xml"):
            # Parse variables from dbgap data dictionary xml file
            variables = annotator.load_data_dictionary(args.crawl_file)
            concepts = annotator.annotate(variables)

        # Add Synonyms - this is slow: 30 secs
        concepts = annotator.add_synonyms_to_identifiers(concepts)

        # Add concept actions (i.e. external links). ATM, it makes a link to the ontology identifier PURL
        # Could be cooler stuff in the future but this'll do for now, pig.
        for concept in concepts:
            ontology_purl = BioLinkPURLerizer.get_curie_purl(concept)
            concepts[concept]["concept_action"] = ontology_purl if ontology_purl is not None else ""

        ''' New - Clean up after POC'''
        # Clean concepts and add ontology descriptors - this is slow: 60 secs
        concepts = annotator.clean_concepts(concepts)

        # TODO: Add more identifiers to variables based on Topmed tag in identifiers slot, for Proof of concept.
        # ##### Not sure if this is what we want to do.  Maybe delete this chunk after POC.
        for variable in variables:
            for identifier in variable['identifiers']:
                if identifier.startswith("TOPMED"):# if TOPMED identifier, expand
                    new_vars = list(concepts[identifier]['identifiers'].keys())
                    variable['identifiers'].extend(new_vars)

            # Make unique
            variable['identifiers'] = list(set(list(variable['identifiers'])))

        variable_file = open("variable_file.json", "w")
        variable_file.write(f"{json.dumps(variables, indent=2)}")
        ''' End New for POC '''

        source = "/graph/gamma/quick" #TODO: NOV 25 - Since we have synonyms, we can change this to /schema, probably.
        queries = {
            "disease": tql.QueryFactory(["disease", "phenotypic_feature"], source),
            "pheno": tql.QueryFactory(["phenotypic_feature", "disease"], source),
            "anat": tql.QueryFactory(["disease", "anatomical_entity"], source),
            "chem_to_disease": tql.QueryFactory(["chemical_substance", "disease"], source),
            "phen_to_anat": tql.QueryFactory(["phenotypic_feature", "anatomical_entity"], source),
            "anat_to_disease": tql.QueryFactory(["anatomical_entity", "disease"], source),
            "anat_to_pheno": tql.QueryFactory(["anatomical_entity", "phenotypic_feature"], source)
        }

        # List of identifiers to stay away from for now
        query_exclude_identifiers = ["CHEBI:17336"]

        if args.crawl_by_tag:
            # Append tag info to variables
            search.crawl_by_tag(tags,
                                variables,
                                tag_index,
                                kg_index,
                                queries,
                                min_score=args.min_tranql_score,
                                query_exclude_identifiers=query_exclude_identifiers)

        elif args.crawl:
            search.crawl(tags,
                         variables,
                         tag_index,
                         queries,
                         min_score=args.min_tranql_score,
                         query_exclude_identifiers=query_exclude_identifiers)
        
        elif args.crawl_by_concept:
            search.crawl_by_concept(
                concepts,
                concepts_index,
                kg_index,
                queries,
                min_score=args.min_tranql_score,
                query_exclude_identifiers=query_exclude_identifiers)

            # TODO: Index the Variables separately (this doesn't require crawling)
            search.index_variables(variables,variables_index)



import argparse
import json
import logging
import os
from dug.core import Search
from dug.annotate import TOPMedStudyAnnotator, GraphDB
from typing import Dict

logger = logging.getLogger (__name__)

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

Config = Dict

def main ():
    """
    Configure an annotator.
    """
    db_url_default = "http://" + os.environ.get('NEO4J_HOST', 'localhost') + ":" + os.environ.get('NEO4J_PORT', '7474') + "/db/data"
    config = {
        'annotator'      : "https://api.monarchinitiative.org/api/nlp/annotate/entities?min_length=4&longest_only=false&include_abbreviation=false&include_acronym=false&include_numbers=false&content=",
        'normalizer'     : "https://nodenormalization-sri.renci.org/get_normalized_nodes?curie=",
        'synonym_service': "https://onto.renci.org/synonyms/",
        'ontology_metadata': "https://api.monarchinitiative.org/api/ontology/term/",
        'password'       : os.environ['NEO4J_PASSWORD'],
        'username'       : 'neo4j',
        'db_url'         : db_url_default,
        'redis_host'     : os.environ.get('REDIS_HOST', 'localhost'),
        'redis_port'     : os.environ.get('REDIS_PORT', 6379),
        'redis_password' : os.environ.get('REDIS_PASSWORD', ''),
    }

    parser = argparse.ArgumentParser(description='Load edges and nodes into Neo4j via kgx')
    parser.add_argument('--load', help='annotation file to load via kgx')
    parser.add_argument('--annotate', help='annotate TOPMed data dictionary file', default=None)
    parser.add_argument('--db-url', help='database url', default=db_url_default)
    parser.add_argument('--db-username', help='database username', default='neo4j')
    parser.add_argument('--db-password', help='database password', default=os.environ['NEO4J_PASSWORD'])
    parser.add_argument('--tagged', help='annotate tagged variables')
    parser.add_argument('--index', help='build search index directly. valid only with --tagged.', default=False)
    parser.add_argument('--debug', help='debug output', default=False)
    args = parser.parse_args()

    config['username'] = args.db_username
    config['password'] = args.db_password
    config['db_url']   = args.db_url
    annotator = TOPMedStudyAnnotator (config=config)

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    if args.load:
        """ Load an annotated data dictionary and write to a graph database. """
        with open(args.load, 'r') as stream:
            obj = json.load(stream)

            """ Convert. """
            knowledge_graph = annotator.convert_to_kgx_json (obj)

            """ Write output """
            annotator.write (knowledge_graph)
    elif args.annotate:

        input_file = args.annotate
        variables = annotator.load_data_dictionary (input_file) \
                    if input_file.endswith('.xml') \
                       else annotator.load_csv (input_file)

        """ Annotate an input file using the MonarchInitiative SciGraph named entity extractor and annotator. """
        response = annotator.annotate (variables)

        """ Write the annotated variables. """
        output_file = args.annotate.\
                      replace ('.xml', '_tagged.json').\
                      replace ('.csv', '_tagged.json')
        with open(output_file, 'w') as stream:
            json.dump(response, stream, indent=2)
    elif args.tagged:

        """
        The link operator for tagged variables.
        This annotates, normalizes, converts to a kgs graph, and loads via KGX into Neo4J.
        """
        variables, tags = annotator.load_tagged_variables (args.tagged)
        tags = annotator.annotate (tags)
        logger.debug (f"{json.dumps(tags, indent=2)}")
        knowledge_graph = annotator.make_tagged_kg (variables, tags)
        logger.debug (f"{json.dumps(knowledge_graph, indent=2)}")
        annotator.write (knowledge_graph)

        if (args.index):
            """ PHASE-1: Go directly to Elasticsearch, bypassing Translator. """

            """ Initialize the search engine core. """
            search = Search ()

            """ Connect to the graph database. """
            graph = GraphDB (config)

            """
            For each Biolink model category, get subgraphs initiated with the category, connected to a tag, to a variable, to a study.
            For each resulting subgraph, build a document linking data from each of these layers.
            Add the document to the search index.
            """
            for category in [ "phenotypic_feature", "anatomical_entity", "cell_type", "gene", "disease", "biological_process" ]:
                page_size= 10_000
                path_count_cypher = f"""MATCH 
                    p=(bioconcept:{category})--(tag:information_content_entity)--(variable:clinical_modifier)--(study:clinical_trial) 
                    RETURN count(p) as path_count
                """
                get_path_query = lambda skip, limit: f"MATCH" \
                    f"(bioconcept:{category})--(tag:information_content_entity)--(variable:clinical_modifier)--(study:clinical_trial)" \
                    f"RETURN * SKIP {skip} LIMIT {limit}"
                path_count = graph.query(path_count_cypher)[0][0]
                logger.debug(f"Found {path_count} paths for ({category})--(information_content_entity)"
                             "--(clinical_modifier)--(clinical_trial)")
                for offset in range(0, path_count, page_size):
                    skip = offset
                    limit = skip + page_size
                    query = get_path_query(skip, limit)
                    result = graph.query (query)
                    if result:
                        for index, row in enumerate(result.rows):
                            """
                            We're looking at a single subgraph returned from the query.
                            Build a document collecting these artifacts.
                            """

                            """ 
                            Initialize empty dict with initial values 
                            set to null so that they'll always exist 
                            """
                            doc = {'study_name': '',
                                   'description': '',
                                   'instructions': ''}

                            for node in row:
                                node_id = node['id']
                                if node_id.startswith ("TOPMED.VAR:"):
                                    doc['var'] = node_id.replace  ("TOPMED.VAR:", "")
                                elif node_id.startswith ("TOPMED.STUDY"):
                                    doc['study'] = node_id.replace ("TOPMED.STUDY:", "")
                                    doc['study_name'] = node['name']
                                elif node_id.startswith ("TOPMED.TAG"):
                                    doc['tag'] = node_id.replace ("TOPMED.TAG:", "")
                                    doc['description'] = node['description']
                                    doc['instructions'] = node['instructions']
                                else:
                                    doc['id'] = node_id
                                    doc['name'] = node['name']

                            unique_doc_id = f"{doc['id']}_{doc['study']}_{doc['var']}"
                            logger.debug (f"{json.dumps(doc, indent=2)}")

                            """ Index the document. """
                            search.index_doc (
                                index='test',
                                doc=doc,
                                doc_id=unique_doc_id)


if __name__ == '__main__':
    main ()



import csv
import json
import logging
import os
import redis
import traceback
import re
import urllib
import xml.etree.ElementTree as ET
#from kgx import NeoTransformer, JsonTransformer
from neo4jrestclient.client import GraphDatabase
import requests
from requests_cache import CachedSession
from typing import List, Dict
import hashlib

logger = logging.getLogger (__name__)

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

Config = Dict


def get_dbgap_var_link(study_id, variable_id):
    base_url = "https://www.ncbi.nlm.nih.gov/projects/gap/cgi-bin/variable.cgi"
    return f'{base_url}?study_id={study_id}&phv={variable_id}'


def get_dbgap_study_link(study_id):
    base_url = "https://www.ncbi.nlm.nih.gov/projects/gap/cgi-bin/study.cgi"
    return f'{base_url}?study_id={study_id}'


def parse_study_name_from_filename(filename):
    # Parse the study name from the xml filename if it exists. Return None if filename isn't right format to get id from
    dbgap_file_pattern = re.compile(r'.*/*phs[0-9]+\.v[0-9]\.pht[0-9]+\.v[0-9]\.(.+)\.data_dict.*')
    match = re.match(dbgap_file_pattern, filename)
    if match is not None:
        return match.group(1)
    return None


class Debreviator:
    """ Expand certain abbreviations to increase our hit rate."""
    def __init__(self):
        self.decoder = {
            "bmi" : "body mass index"
        }

    def decode(self, text):
        for key, value in self.decoder.items():
            text = text.replace(key, value)
        return text


class TOPMedStudyAnnotator:
    """
    Annotate TOPMed study data with semantic knowledge graph linkages.

    """
    def __init__(self, config: Config):
        self.normalizer = config['normalizer']
        self.annotator = config['annotator']
        self.synonym_service = config['synonym_service']
        self.ontology_metadata = config['ontology_metadata']
        self.db_url = config['db_url']
        self.username = config['username']
        self.password = config['password']
        self.redis_host = config['redis_host']
        self.redis_port = config['redis_port']
        self.redis_password = config['redis_password']
        self.debreviator = Debreviator ()

    def load_data_dictionary (self, input_file : str) -> Dict:
        """
        This loads a data dictionary. It's unclear if this will be useful going forwar. But for now,
        it demonstrates how much of the rest of the  pipeline might be applied to data dictionaries.
        """
        tree = ET.parse(input_file)
        root = tree.getroot()
        study_id = root.attrib['study_id']
        participant_set = root.attrib['participant_set']

        # Parse study name from filehandle
        study_name = parse_study_name_from_filename(input_file)
        if study_name is None:
            err_msg = f"Unable to parse DbGaP study name from data dictionary: {input_file}!"
            logger.error(err_msg)
            raise IOError(err_msg)

        variables = [{
            "element_id"        : f"{variable.attrib['id']}.p{participant_set}",
            "element_name"      : variable.find ('name').text,
            "element_desc"      : variable.find ('description').text.lower(),
            "collection_id"     : f"{study_id}.p{participant_set}",
            "collection_name"   : study_name,
            "collection_desc"   : "",
            "data_type": "dbGap_variable",
            "identifiers": {}
        } for variable in root.iter('variable')]

        # Create DBGaP links as study/variable actions
        for variable in variables:
            variable["collection_action"] = get_dbgap_study_link(study_id = variable["collection_id"])
            variable["element_action"] = get_dbgap_var_link(study_id = variable["collection_id"],
                                                               variable_id=variable["element_id"].split(".")[0].split("phv")[1])
        return variables

    def load_tagged_variables (self, input_file : str) -> Dict:
        """
        Load tagged variables.
          Presumes a harmonized variable list as a CSV file as input.
          A naming convention such that <prefix>_variables_<version>.csv will be the form of the filename.
          An adjacent file called <prefix>_tags_<version.json will exist.

          :param input_file: A list of study variables linked to tags, studies, and other relevant data.
          :returns: Returns variables, a list of parsed variable dictionaries and tags, a list of parsed
                    tag definitions linked to the variabels.
        """
        tags_input_file = input_file.replace (".csv", ".json").replace ("_variables_", "_tags_")
        if not os.path.exists (tags_input_file):
            raise ValueError (f"Accompanying tags file: {tags_input_file} must exist.")
        variables = []
        headers = "tag_pk 	tag_title 	variable_phv 	variable_full_accession 	dataset_full_accession 	study_full_accession 	study_name 	study_phs 	study_version 	created 	modified"
        with open(input_file, newline='') as csvfile:
            reader = csv.DictReader(csvfile, delimiter='\t',)
            for row in reader:
                better_row = { k.strip () : v for  k, v in row.items () }
                row = better_row
                logger.debug(f"{json.dumps(row, indent=2)}")
                variables.append ({
                    "element_id"                     : f"{row['variable_full_accession']}",
                    "element_name"                   : row['variable_name'] if 'variable_name' in row else row['variable_full_accession'],
                    "element_desc"            : row['variable_desc'] if 'variable_name' in row else row['variable_full_accession'],
                    "collection_id"               : f"{row['study_full_accession']}",
                    "collection_name"             : row['study_name'],
                    "collection_desc"      : "",
                    "data_type": "dbGap_variable",
                    "identifiers": [f"TOPMED.TAG:{row['tag_pk']}"]
                })
                logger.debug(f"{json.dumps(variables[-1], indent=2)}")

        # Create DBGaP links as study/variable actions
        for variable in variables:
            variable["element_action"] = get_dbgap_study_link(study_id=variable["collection_id"])
            variable["collection_action"] = get_dbgap_var_link(study_id=variable["collection_id"],
                                                               variable_id=variable["element_id"])
        # Create concepts for each tag
        with open(tags_input_file, "r") as stream:
            tags = json.load (stream)
        for tag in tags:
            for f, v in tag['fields'].items ():
                tag[f] = v
            del tag[f]
            tag['id'] = f"TOPMED.TAG:{tag['pk']}"
            tag['identifiers'] = {}
            tag['is_variable_tag'] = True
            tag['type'] = 'TOPMed Phenotype Concept'

        return variables, tags

    def normalize(self, http_session, curie, url, variable) -> None:
        """ Given an identifier (curie), use the Translator SRI node normalization service to
            find a preferred identifier, equivalent identifiers, and biolink model types for the node.

            :param http_session: A requests session to use for HTTP requests.
            :param curie: The identifier to normalize.
            :param url: The URL to use.
            :param variable: The variable in which to record the normalized identifier.
        """
        """
        Added blank_preferred_id to keep terms in the KG that fail to normalize.
        """
        blank_preferred_id = {
            "label": "",
            "equivalent_identifiers": [],
            "type": ['named_thing'] # Default NeoTransformer category
        }
        """ Normalize the identifier with respect to the BioLink Model. """

        logger.debug(f"Normalizing: {curie}")
        normalized = http_session.get(url).json ()

        """ Record normalized results. """
        normalization = normalized.get(curie, {})
        if normalization is None:
            variable['identifiers'][curie] = blank_preferred_id
            logger.error (f"Normalization service did not return normalization for: {curie}")
            return

        preferred_id = normalization.get ("id", {})
        equivalent_identifiers = normalization.get ("equivalent_identifiers", [])
        biolink_type = normalization.get ("type", [])

        """ Build the response. """
        if 'identifier' in preferred_id:
            logger.debug(f"Preferred id: {preferred_id}")
            variable['identifiers'][preferred_id['identifier']] = {
                "label" : preferred_id.get('label',''),
                "equivalent_identifiers" : [ v['identifier'] for v in equivalent_identifiers ],
                "type" : biolink_type
            }
        else:
            variable['identifiers'][curie] = blank_preferred_id
            logger.debug (f"ERROR: normaliz({curie})=>({preferred_id}). No identifier?")

    def annotate (self, variables : Dict) -> Dict:
        """
        This operates on a dbGaP data dictionary which is
          an XML formatted study with a data_table root element containing a list of variables.
        - Tag prose variable descriptions with ontology identifiers.
        - Resolve those identifiers via Translator to preferred identifiers and BioLink Model categories.
        Specifically, for each variable
          create an object to represent the variable
          perform NLP annotation of the variable based on its description
            use the Monarch NLP annotator with named entity recognition
            (add additional NLP, eg. Vanderbilt and others here...?)
          normalize the resulting identifiers using the Translator normalization API

          :param variables: A dictionary of variables.
          :returns: A dictionary of annotated variables.
        """

        """
        Initialize and reuse a cached HTTP session for more efficient connection management.
        Use the Redis backend for requests-cache to store results accross executions
        """
        redis_connection = redis.StrictRedis (host=self.redis_host,
                                              port=self.redis_port,
                                              password=self.redis_password)
        http_session = CachedSession (
            cache_name='annotator',
            backend="redis",
            connection=redis_connection)

        variable_file = open("normalized_inputs.txt", "w")

        # Initialize return dict
        concepts = {}
        norm_fails = open("norm_fails.txt", "w")

        """ Annotate and normalize each tag. """
        for variable in variables:
            logger.debug(variable)
            try:
                """
                If the variable has an Xref identifier, normalize it.
                This data is only in the CSV formatted harmonized variable data.
                If that format goes away, delete this.
                """
                if 'xref' in variable:
                    self.normalize (http_session,
                                    variable['xref'],
                                    f"{self.normalizer}{variable['xref']}",
                                    variable)

                """ Annotate ontology terms in the text. """
                if 'element_desc' not in variable and 'description' not in variable:
                    logger.warn (f"this variable has no description: {json.dumps(variable, indent=2)}")
                    continue

                # Use different desc field based on whether we're dealing with a variable tag vs. just a normal variable
                description_field = "description" if "is_variable_tag" in variable else "element_desc"
                description = variable[description_field].replace ("_", " ")
                description = self.debreviator.decode (description)

                # Annotation
                encoded = urllib.parse.quote (description)
                url = f"{self.annotator}{encoded}"
                annotations = http_session.get(url).json ()

                """ Normalize each ontology identifier from the annotation. """
                for span in annotations.get('spans',[]):
                    search_text = span.get('text',None) # Always a string
                    for token in span.get('token',[]):
                        curie = token.get('id', None)
                        if not curie:
                            continue
                        self.normalize (http_session,
                                        curie,
                                        f"{self.normalizer}{curie}",
                                        variable)

                        # Skip if failing normalization
                        #TODO Address this failing normalization.
                        if curie not in variable['identifiers']:
                            norm_fails.write(f"{curie}\n")
                            continue

                        # Add concepts
                        term = token.get('terms', None) # Always a list
                        if curie not in concepts:
                            concepts[curie] = {
                                "id": curie,
                                "name": term[0],
                                "description": "",
                                "type": "",
                                "search_terms": [search_text] + term,
                                "identifiers": {
                                    curie: variable['identifiers'][curie]
                                }
                            }
                        else:
                            concepts[curie]["search_terms"].extend([search_text] + term)

            except json.decoder.JSONDecodeError as e:
                traceback.print_exc ()
            except:
                traceback.print_exc ()
                raise
            variable_file.write(f"{json.dumps(variable, indent=2)}\n")

            # Optionally create a concept when variable is actually a pre-harmonized variable tag (e.g. TOPMed tags)
            if "is_variable_tag" in variable:
                concepts[variable['id']] = {
                    "id": variable['id'],
                    "name": variable['title'],
                    "description": f"{description}. {variable['instructions']}",
                    "type": variable["type"],
                    "search_terms": [],
                    "identifiers" : variable['identifiers']
                }

        return concepts

    def write (self, graph : Dict) -> None:
        """
        Given a dictionary that is a graph containing nodes edges,
        use KGX to load the graph into a Neo4J database.

        :param graph: A KGX formatted graph as dictionary.
        """

        """ Load the knowledge graph into KGX and emit to Neo4J. """
        json_transformer = JsonTransformer ()
        json_transformer.load (graph)
        db = NeoTransformer (json_transformer.graph,
                             self.db_url,
                             self.username,
                             self.password)
        db.save()
        db.neo4j_report()

    def make_edge (self,
                   subj : str,
                   pred : str,
                   obj  : str,
                   edge_label : str ='association',
                   category : List[str] = []
    ):
        """
        Create an edge between two nodes.

        :param subj: The identifier of the subject.
        :param pred: The predicate linking the subject and object.
        :param obj: The object of the relation.
        :param edge_label: Label for the edge.
        :param category: The list of Biolink categories relating to the edge.
        :returns: Returns and edge.
        """
        edge_id = hashlib.md5(f'{subj}{pred}{obj}'.encode('utf-8')).hexdigest()
        return {
            "subject"     : subj,
            "predicate"   : pred,
            "id": edge_id,
            "edge_label"  : edge_label if len(edge_label) > 0 else "n/a",
            "object"      : obj,
            "provided_by" : "renci.bdc.semanticsearch.annotator",
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#type": "OBAN:association",
            "category" : category
        }

    def make_tagged_kg (self, variables : Dict, tags : Dict) -> Dict:
        """ Make a Translator standard knowledge graph representing
        tagged study variables.
        :param variables: The variables to model.
        :param tags: The tags characterizing the variables.
        :returns: Returns dictionary with nodes and edges modeling a Translator/Biolink KG.
        """
        graph = {
            "nodes" : [],
            "edges" : []
        }
        edges = graph['edges']
        nodes = graph['nodes']
        studies = {}
        tag_ref = {}

        """ Create graph elements to model tags and their
        links to identifiers gathered by semantic tagging. """
        tag_map = {}
        for tag in tags:
            tag_pk = tag['pk']
            tag_id  = tag['id']
            tag_map[tag_pk] = tag
            nodes.append ({
                "id" : tag_id,
                "pk" : tag_pk,
                "name" : tag['title'],
                "description" : tag['description'],
                "instructions" : tag['instructions'],
                "category" : [ "information_content_entity" ]
            })
            """ Link ontology identifiers we've found for this tag via nlp. """
            for identifier, metadata in tag['identifiers'].items ():
                nodes.append ({
                    "id" : identifier,
                    "name" : metadata['label'],
                    "category" : metadata['type']
                })
                edges.append (self.make_edge (
                    subj=tag_id,
                    pred="OBO:RO_0002434",
                    obj=identifier,
                    edge_label='association',
                    category=[ "association" ]))
                edges.append (self.make_edge (
                    subj=identifier,
                    pred="OBO:RO_0002434",
                    obj=tag_id,
                    edge_label='association',
                    category=[ "association" ]))

        """ Create nodes and edges to model variables, studies, and their
        relationships to tags. """
        for variable in variables:
            variable_id = variable['variable_id']
            variable_tag_pk = variable['tag_pk']
            study_id = variable['study_id']
            tag_id = tag_map[int(variable_tag_pk)]['id']
            if not study_id in studies:
                nodes.append ({
                    "id" : study_id,
                    "name" : variable['study_name'],
                    "category" : [ "clinical_trial" ]
                })
                studies[study_id] = study_id

            nodes.append ({
                "id" : variable_id,
                "name" : variable_id,
                "category" : [ "clinical_modifier" ]
            })
            """ Link to its study.  """
            edges.append (self.make_edge (
                subj=variable_id,
                edge_label='part_of',
                pred="OBO:RO_0002434",
                obj=study_id,
                category=['part_of']))
            edges.append (self.make_edge (
                subj=study_id,
                edge_label='has_part',
                pred="OBO:RO_0002434",
                obj=variable_id,
                category=['has_part']))

            """ Link to its tag. """
            edges.append (self.make_edge (
                subj=variable_id,
                edge_label='part_of',
                pred="OBO:RO_0002434",
                obj=tag_id,
                category=['part_of']))
            edges.append (self.make_edge (
                subj=tag_id,
                edge_label='has_part',
                pred="OBO:RO_0002434",
                obj=variable_id,
                category=['has_part']))

        return graph

    def convert_to_kgx_json (self, annotations):
        """
        Given an annotated and normalized set of study variables,
        generate a KGX compliant graph given the normalized annotations.
        Write that grpah to a graph database.
        See BioLink Model for category descriptions. https://biolink.github.io/biolink-model/notes.html
        """
        graph = {
            "nodes" : [],
            "edges" : []
        }
        edges = graph['edges']
        nodes = graph['nodes']

        for index, variable in enumerate(annotations):
            study_id = variable['study_id']
            if index == 0:
                """ assumes one study in this set. """
                nodes.append ({
                    "id" : study_id,
                    "category" : [ "clinical_trial" ]
                })

            """ connect the study and the variable. """
            edges.append (self.make_edge (
                subj=variable['variable_id'],
                edge_label='part_of',
                pred="OBO:RO_0002434",
                obj=study_id,
                category=['part_of']))
            edges.append (self.make_edge (
                subj=study_id,
                edge_label='has_part',
                pred="OBO:RO_0002434",
                obj=variable['variable_id'],
                category=['has_part']))

            """ a node for the variable. """
            nodes.append ({
                "id" : variable['variable_id'],
                "name" : variable['variable'],
                "description" : variable['description'],
                "category" : [ "clinical_modifier" ]
            })
            for identifier, metadata in variable['identifiers'].items ():
                edges.append (self.make_edge (
                    subj=variable['variable_id'],
                    pred="OBO:RO_0002434",
                    obj=identifier,
                    edge_label='association',
                    category=[ "case_to_phenotypic_feature_association" ]))
                edges.append (self.make_edge (
                    subj=identifier,
                    pred="OBO:RO_0002434",
                    obj=variable['variable_id'],
                    edge_label='association',
                    category=[ "case_to_phenotypic_feature_association" ]))
                nodes.append ({
                    "id" : identifier,
                    "name" : metadata['label'],
                    "category" : metadata['type']
                })
        return graph

    def get_variables_from_tranql(self):
        '''
        This function returns the tagged variables,
        including identifier information,
        from TranQL.  It is the starting point for
        indexing documents into ElasticSearch
        '''

        # Create a session and connect to TranQL
        tranql_endpoint = "https://tranql.renci.org/tranql/query?dynamic_id_resolution=true&asynchronous=false"
        headers = {
            "accept" : "application/json",
            "Content-Type" : "text/plain"
        }
        s = requests.Session()
        s.headers.update(headers)
        #TODO: Clean up the session object by bringing header, url, and data into attributes of obj

        # Build TranQL Query
        source = "/schema"
        questions = ["information_content_entity",
                     "clinical_modifier",
                     "clinical_trial"]
        query = f'select {"->".join(questions)} from "{source}"'
        single_query = query + f" where information_content_entity = 'TOPMED.TAG:51'"

        # Put this in while loop to iterate until we are successful in connecting to TranQL
        while True:
            try:
                response = s.post(url = tranql_endpoint,
                            data = query)
                if response.status_code is not 200:
                    logging.error(f"Encountered error: {response.status_code}:{response.reason}.  Trying again...")
                    continue
                break
            except requests.ConnectionError as e:
                logger.error (f"Encountered exception: {e}.  Trying again...")

        # Build List of Dicts
        response = response.json () # transform to JSON document if successful.
        nodes = response['knowledge_graph']['nodes']
        edges = response['knowledge_graph']['edges']
        tags = list(filter(lambda x: x['id'].startswith("TOPMED.TAG"),nodes))
        variables = list(filter(lambda x: x['id'].startswith("TOPMED.VAR"),nodes))
        studies = list(filter(lambda x: x['id'].startswith("TOPMED.STUDY"),nodes))

        # Annotate Tags w/ identifiers
        identifier_type = "biological_entity"
        
        for tag in tags:
            tag['title'] = tag.pop('name')
            # TranQL query to identifiers
            logger.debug(f"Querying Tag: {tag['title']} with query: {query}")
            query = f'select {identifier_type}->information_content_entity from "{source}" where information_content_entity = "{tag["id"]}"'
            response = s.post(url = tranql_endpoint,
                            data = query).json()
            logger.debug(response)
            identifiers = [identifier for identifier in response['knowledge_graph']['nodes'] if identifier['id'] != tag['id']]
            tag['identifiers'] = {}
            for identifier in identifiers:
                try:
                    # change 'name' to 'label' for downstream functionality.
                    identifier['label'] = identifier.pop('name') 
                except KeyError as e:
                    logger.error(f"{e}: Identifier does not have name/label.  Adding blank label")
                    identifier['label'] = ""
                tag['identifiers'][identifier['id']] = identifier
            '''
            tag_edges = [edge for edge in edges if edge['source_id']==tag['id'] and not edge['target_id'].startswith('TOPMED.VAR')]
            identifiers = [tag['target_id'] for tag in tag_edges]
            for identifier in identifiers:
                # Find identifiers in node
                value = [node for node in nodes if node['id'] == identifier][0] # finds single dictionary
                if 'label' not in value:
                    value['label'] = value.pop('name') # change 'name' to 'label' for downstream processing
                tag['identifiers'][identifier] = value
            '''

        # Add studies and tag_pk to variables
        for variable in variables:
            # Rename
            variable['variable_id'] = variable.pop('id')

            # Grab from tags
            associations = [edge['source_id'] for edge in edges if edge['target_id']==variable['variable_id']]
            variable['tag_pk'] = next(filter(lambda x: x.startswith('TOPMED.TAG'), associations), "").split(":")[1]
            variable['study_id'] = next(filter(lambda x: x.startswith('TOPMED.STUDY'), associations), "")

            # Grab from studies - only linked to one study name
            variable['study_name'] = [study['name'] for study in studies if study['id']==variable['study_id']][0]

        # Return tags and variables
        return variables, tags

    def add_synonyms_to_identifiers(self, concepts):
        '''
        This function does the following:
        - Initialize http_session for NCATS synonym service
        - list comprehension of all identifiers attached to tags
        - Add synonyms to all tags
        '''
        # Create an http_session like in annotate()
        redis_connection = redis.StrictRedis (host=self.redis_host,
                                              port=self.redis_port,
                                              password=self.redis_password)
        http_session = CachedSession (
            cache_name='annotator',
            backend="redis",
            connection=redis_connection)

        # Go through identifiers in tags
        for concept in concepts:
            for identifier in list(concepts[concept]['identifiers'].keys()):
                try:
                    # Get response from synonym service
                    encoded = urllib.parse.quote (identifier)
                    url = f"{self.synonym_service}{encoded}"
                    raw_synonyms = http_session.get(url).json ()

                    # List comprehension for synonyms
                    synonyms = [synonym['desc'] for synonym in raw_synonyms]
                    
                    # Add to identifier
                    concepts[concept]['identifiers'][identifier]['synonyms'] = synonyms
                    
                    # Add to search terms
                    concepts[concept]['search_terms'] += synonyms
                
                except json.decoder.JSONDecodeError as e:
                    concepts[concept]['identifiers'][identifier]['synonyms'] = []
                    logger.error (f"No synonyms returned for: {identifier}")
        return concepts
    
    def clean_concepts(self, concepts):
        '''
        This function does the following:
        - Make search terms within concepts unique
        - Write concepts to a debug file
        - Add more identifiers to variables for Proof of Concept
        - Add name/description to ontology IDs
        '''
        # Create an http_session like in annotate()
        redis_connection = redis.StrictRedis (host=self.redis_host,
                                              port=self.redis_port,
                                              password=self.redis_password)
        http_session = CachedSession (
            cache_name='annotator',
            backend="redis",
            connection=redis_connection)
        
        # Clean Concepts
        for concept in concepts:
            concepts[concept]['search_terms'] = list(set(list(concepts[concept]['search_terms'])))
        
        # Add Name and description
        for concept in concepts:
            try:
                # Get response from synonym service
                encoded = urllib.parse.quote (concept)
                url = f"{self.ontology_metadata}{encoded}"
                response = http_session.get(url).json ()

                # List comprehension for synonyms
                name = response.get('label','')
                description = '' if not response.get('description',None) else response.get('description','')
                ontology_type = '' if not response.get('category', None) else response.get('category','')[0]
                
                # Add to concept
                if len(name):
                    concepts[concept]['name'] = name
                if len(description):
                    concepts[concept]['description'] = description
                if len(ontology_type):
                    concepts[concept]['type'] = ontology_type
            
            except json.decoder.JSONDecodeError as e:
                logger.error (f"No labels returned for: {concept}")
        
        # Write to file
        concept_file = open("concept_file.json", "w")
        concept_file.write(f"{json.dumps(concepts, indent=2)}")
        
        return concepts


class GraphDB:
    def __init__(self, conf):
        self.conf = conf
        self.gdb = GraphDatabase(url=conf['db_url'],
                                 username=conf['username'],
                                 password=conf['password'])
    def query (self, query):
        return self.gdb.query(query, data_contents=True)



import os
import re
import sys
from shutil import rmtree
from setuptools import setup, find_packages, Command

current = os.path.abspath(os.path.dirname(__file__))
def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]
install_reqs = parse_requirements(os.path.join(current, "requirements.txt"))
requirements = [str(r) for r in install_reqs]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open(os.path.join(current, "dug", "__init__.py"), encoding="utf8") as f:
    version = re.search(r'__version__="(.*?)"', f.read()).group(1)

class PublishClass(Command):
    description = "Publish the package"
    user_options = []

    # This method must be implemented
    def initialize_options(self):
        pass

    # This method must be implemented
    def finalize_options(self):
        pass

    def run(self):
        try:
            print(f"-----> removing previous builds")
            rmtree(os.path.join(current, 'dist'))
            rmtree(os.path.join(current, 'build'))
        except Exception as e:
            print(f"-----> Exception: {e}")
            pass
        os.system('python setup.py sdist bdist_wheel --universal')
        os.system('twine upload dist/*')
        sys.exit()

setup(
    name="dug-test",
    version=version,
    maintainer="Renci",
    maintainer_email="muralikarthik.k@renci.org",
    description="Tycho is an API, compiler, and executor for cloud native distributed systems.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/helxplatform/dug.git",
    packages=['dug'],
    include_package_data=True,
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    cmdclass={
        'publish': PublishClass,
    },
)



class BioLinkPURLerizer:
      # Static class for the sole purpose of doing lookups of different ontology PURLs
      # Is it pretty? No. But it gets the job done.
      biolink_lookup = {"APO": "http://purl.obolibrary.org/obo/APO_",
      "Aeolus": "http://translator.ncats.nih.gov/Aeolus_",
      "BIOGRID": "http://identifiers.org/biogrid/",
      "BIOSAMPLE": "http://identifiers.org/biosample/",
      "BSPO": "http://purl.obolibrary.org/obo/BSPO_",
      "CAID": "http://reg.clinicalgenome.org/redmine/projects/registry/genboree_registry/by_caid?caid=",
      "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
      "CHEMBL.COMPOUND": "http://identifiers.org/chembl.compound/",
      "CHEMBL.MECHANISM": "https://www.ebi.ac.uk/chembl/mechanism/inspect/",
      "CHEMBL.TARGET": "http://identifiers.org/chembl.target/",
      "CID": "http://pubchem.ncbi.nlm.nih.gov/compound/",
      "CL": "http://purl.obolibrary.org/obo/CL_",
      "CLINVAR": "http://identifiers.org/clinvar/",
      "CLO": "http://purl.obolibrary.org/obo/CLO_",
      "COAR_RESOURCE": "http://purl.org/coar/resource_type/",
      "CPT": "https://www.ama-assn.org/practice-management/cpt/",
      "CTD": "http://translator.ncats.nih.gov/CTD_",
      "ClinVarVariant": "http://www.ncbi.nlm.nih.gov/clinvar/variation/",
      "DBSNP": "http://identifiers.org/dbsnp/",
      "DGIdb": "https://www.dgidb.org/interaction_types",
      "DOID": "http://purl.obolibrary.org/obo/DOID_",
      "DRUGBANK": "http://identifiers.org/drugbank/",
      "DrugCentral": "http://translator.ncats.nih.gov/DrugCentral_",
      "EC": "http://www.enzyme-database.org/query.php?ec=",
      "ECTO": "http://purl.obolibrary.org/obo/ECTO_",
      "EDAM-DATA": "http://edamontology.org/data_",
      "EDAM-FORMAT": "http://edamontology.org/format_",
      "EDAM-OPERATION": "http://edamontology.org/operation_",
      "EDAM-TOPIC": "http://edamontology.org/topic_",
      "EFO": "http://identifiers.org/efo/",
      "ENSEMBL": "http://identifiers.org/ensembl/",
      "ExO": "http://purl.obolibrary.org/obo/ExO_",
      "FAO": "http://purl.obolibrary.org/obo/FAO_",
      "FB": "http://identifiers.org/fb/",
      "FBcv": "http://purl.obolibrary.org/obo/FBcv_",
      "FlyBase": "http://flybase.org/reports/",
      "GAMMA": "http://translator.renci.org/GAMMA_",
      "GO": "http://purl.obolibrary.org/obo/GO_",
      "GOLD.META": "http://identifiers.org/gold.meta/",
      "GOP": "http://purl.obolibrary.org/obo/go#",
      "GOREL": "http://purl.obolibrary.org/obo/GOREL_",
      "GSID": "https://scholar.google.com/citations?user=",
      "GTEx": "https://www.gtexportal.org/home/gene/",
      "HANCESTRO": "http://www.ebi.ac.uk/ancestro/ancestro_",
      "HCPCS": "http://purl.bioontology.org/ontology/HCPCS/",
      "HGNC": "http://identifiers.org/hgnc/",
      "HGNC.FAMILY": "http://identifiers.org/hgnc.family/",
      "HMDB": "http://identifiers.org/hmdb/",
      "HP": "http://purl.obolibrary.org/obo/HP_",
      "ICD0": "http://translator.ncats.nih.gov/ICD0_",
      "ICD10": "http://translator.ncats.nih.gov/ICD10_",
      "ICD9": "http://translator.ncats.nih.gov/ICD9_",
      "INCHI": "http://identifiers.org/inchi/",
      "INCHIKEY": "http://identifiers.org/inchikey/",
      "INTACT": "http://identifiers.org/intact/",
      "IUPHAR.FAMILY": "http://identifiers.org/iuphar.family/",
      "KEGG": "http://identifiers.org/kegg/",
      "LOINC": "http://loinc.org/rdf/",
      "MEDDRA": "http://identifiers.org/meddra/",
      "MESH": "http://identifiers.org/mesh/",
      "MGI": "http://identifiers.org/mgi/",
      "MI": "http://purl.obolibrary.org/obo/MI_",
      "MIR": "http://identifiers.org/mir/",
      "MONDO": "http://purl.obolibrary.org/obo/MONDO_",
      "MP": "http://purl.obolibrary.org/obo/MP_",
      "MSigDB": "https://www.gsea-msigdb.org/gsea/msigdb/",
      "MetaCyc": "http://translator.ncats.nih.gov/MetaCyc_",
      "NCBIGENE": "http://identifiers.org/ncbigene/",
      "NCBITaxon": "http://purl.obolibrary.org/obo/NCBITaxon_",
      "NCIT": "http://purl.obolibrary.org/obo/NCIT_",
      "NDDF": "http://purl.bioontology.org/ontology/NDDF/",
      "NLMID": "https://www.ncbi.nlm.nih.gov/nlmcatalog/?term=",
      "OBAN": "http://purl.org/oban/",
      "OBOREL": "http://purl.obolibrary.org/obo/RO_",
      "OIO": "http://www.geneontology.org/formats/oboInOwl#",
      "OMIM": "http://purl.obolibrary.org/obo/OMIM_",
      "ORCID": "https://orcid.org/",
      "ORPHA": "http://www.orpha.net/ORDO/Orphanet_",
      "ORPHANET": "http://identifiers.org/orphanet/",
      "PANTHER.FAMILY": "http://identifiers.org/panther.family/",
      "PANTHER.PATHWAY": "http://identifiers.org/panther.pathway/",
      "PATO-PROPERTY": "http://purl.obolibrary.org/obo/pato#",
      "PDQ": "https://www.cancer.gov/publications/pdq#",
      "PHARMGKB.DRUG": "http://identifiers.org/pharmgkb.drug/",
      "PHARMGKB.PATHWAYS": "http://identifiers.org/pharmgkb.pathways/",
      "PHAROS": "http://pharos.nih.gov",
      "PMID": "http://www.ncbi.nlm.nih.gov/pubmed/",
      "PO": "http://purl.obolibrary.org/obo/PO_",
      "POMBASE": "http://identifiers.org/pombase/",
      "PR": "http://purl.obolibrary.org/obo/PR_",
      "PUBCHEM.COMPOUND": "http://identifiers.org/pubchem.compound/",
      "PUBCHEM.SUBSTANCE": "http://identifiers.org/pubchem.substance/",
      "PathWhiz": "http://smpdb.ca/pathways/#",
      "REACT": "http://www.reactome.org/PathwayBrowser/#/",
      "REPODB": "http://apps.chiragjpgroup.org/repoDB/",
      "RGD": "http://identifiers.org/rgd/",
      "RHEA": "http://identifiers.org/rhea/",
      "RNACENTRAL": "http://identifiers.org/rnacentral/",
      "RO": "http://purl.obolibrary.org/obo/RO_",
      "RTXKG1": "http://kg1endpoint.rtx.ai/",
      "RXNORM": "http://purl.bioontology.org/ontology/RXNORM/",
      "ResearchID": "https://publons.com/researcher/",
      "SEMMEDDB": "https://skr3.nlm.nih.gov/SemMedDB",
      "SGD": "http://identifiers.org/sgd/",
      "SIO": "http://semanticscience.org/resource/SIO_",
      "SMPDB": "http://identifiers.org/smpdb/",
      "SNOMEDCT": "http://identifiers.org/snomedct/",
      "SNPEFF": "http://translator.ncats.nih.gov/SNPEFF_",
      "ScopusID": "https://www.scopus.com/authid/detail.uri?authorId=",
      "TAXRANK": "http://purl.obolibrary.org/obo/TAXRANK_",
      "UBERGRAPH": "http://translator.renci.org/ubergraph-axioms.ofn#",
      "UBERON": "http://purl.obolibrary.org/obo/UBERON_",
      "UBERON_CORE": "http://purl.obolibrary.org/obo/uberon/core#",
      "UMLS": "http://identifiers.org/umls/",
      "UMLSSC": "https://metamap.nlm.nih.gov/Docs/SemanticTypes_2018AB.txt/code#",
      "UMLSSG": "https://metamap.nlm.nih.gov/Docs/SemGroups_2018.txt/group#",
      "UMLSST": "https://metamap.nlm.nih.gov/Docs/SemanticTypes_2018AB.txt/type#",
      "UNII": "http://identifiers.org/unii/",
      "UPHENO": "http://purl.obolibrary.org/obo/UPHENO_",
      "UniProtKB": "http://identifiers.org/uniprot/",
      "VANDF": "https://www.nlm.nih.gov/research/umls/sourcereleasedocs/current/VANDF/",
      "VMC": "https://github.com/ga4gh/vr-spec/",
      "WB": "http://identifiers.org/wb/",
      "WBPhenotype": "http://purl.obolibrary.org/obo/WBPhenotype_",
      "WBVocab": "http://bio2rdf.org/wormbase_vocabulary",
      "WIKIDATA": "https://www.wikidata.org/wiki/",
      "WIKIDATA_PROPERTY": "https://www.wikidata.org/wiki/Property:",
      "WIKIPATHWAYS": "http://identifiers.org/wikipathways/",
      "WormBase": "https://www.wormbase.org/get?name=",
      "ZFIN": "http://identifiers.org/zfin/",
      "ZP": "http://purl.obolibrary.org/obo/ZP_",
      "alliancegenome": "https://www.alliancegenome.org/",
      "biolink": "https://w3id.org/biolink/vocab/",
      "biolinkml": "https://w3id.org/biolink/biolinkml/",
      "chembio": "http://translator.ncats.nih.gov/chembio_",
      "dcterms": "http://purl.org/dc/terms/",
      "dictyBase": "http://dictybase.org/gene/",
      "doi": "https://doi.org/",
      "fabio": "http://purl.org/spar/fabio/",
      "foaf": "http://xmlns.com/foaf/0.1/",
      "foodb.compound": "http://foodb.ca/compounds/",
      "gff3": "https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md#",
      "gpi": "https://github.com/geneontology/go-annotation/blob/master/specs/gpad-gpi-2-0.md#",
      "gtpo": "https://rdf.guidetopharmacology.org/ns/gtpo#",
      "hetio": "http://translator.ncats.nih.gov/hetio_",
      "interpro": "https://www.ebi.ac.uk/interpro/entry/",
      "isbn": "https://www.isbn-international.org/identifier/",
      "isni": "https://isni.org/isni/",
      "issn": "https://portal.issn.org/resource/ISSN/",
      "medgen": "https://www.ncbi.nlm.nih.gov/medgen/",
      "oboformat": "http://www.geneontology.org/formats/oboInOWL#",
      "pav": "http://purl.org/pav/",
      "prov": "http://www.w3.org/ns/prov#",
      "qud": "http://qudt.org/1.1/schema/qudt#",
      "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
      "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
      "skos": "https://www.w3.org/TR/skos-reference/#",
      "wgs": "http://www.w3.org/2003/01/geo/wgs84_pos",
      "xsd": "http://www.w3.org/2001/XMLSchema#",
      "@vocab": "https://w3id.org/biolink/vocab/"}

      @staticmethod
      def get_curie_purl(curie):
            # Split into prefix and suffix
            suffix = curie.split(":")[1]
            prefix = curie.split(":")[0]

            # Check to see if the prefix exists in the hash
            if prefix not in BioLinkPURLerizer.biolink_lookup:
                  return None

            return f"{BioLinkPURLerizer.biolink_lookup[prefix]}{suffix}"

